// SOURCE CODE HEADER
//
// NTService.cpp
// Purpose: Run an application as a NT/95/98 service
// Contact: jeffg@cips.ca
// Legalese: This source code is public domain
// Based on MSDN example code by Nigel Thomson (nigelt@microsoft.com)
// Modified to also run as Win95/98 service by Steve Porter (steve@dcypher.net)
// Modified for the ECDL2K-108 challenge by Jeff Gilchrist (jeffg@cips.ca).

// Implementation of CNTService

#include <windows.h>
#include <stdio.h>

#include "NTService.h"
#include "osinfo.h"

//extern DWORD RegisterServiceProcess(DWORD dwProcessId, DWORD dwType);

typedef DWORD (WINAPI *fp_RegServProc)(DWORD dwProcessId,DWORD dwType); 
fp_RegServProc fncptr = NULL; 

CNTService* CNTService::m_pThis = NULL;

CNTService::CNTService(const char* szServiceName)
{
    // copy the address of the current object so we can access it from
    // the static member callback functions.  WARNING: This limits the 
	// application to only one CNTService object. 
    
	m_pThis = this;
    
    // Set the default service name and version
    
	strncpy(m_szServiceName, szServiceName, sizeof(m_szServiceName) - 1);
    m_iMajorVersion = V_MAJOR; m_iMinorVersion = V_MINOR; m_iBuildVersion = V_BUILD;
    m_hEventSource = NULL; isManual = FALSE;

    // set up the initial service status 
    
	m_hServiceStatus = NULL;
    m_Status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    m_Status.dwCurrentState = SERVICE_STOPPED;
    m_Status.dwControlsAccepted = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
    m_Status.dwWin32ExitCode = 0;
    m_Status.dwServiceSpecificExitCode = 0;
    m_Status.dwCheckPoint = 0;
    m_Status.dwWaitHint = 0;
    m_bIsRunning = FALSE;
}

CNTService::~CNTService(void)
{
    DebugMsg("CNTService::~CNTService()");

    if(m_hEventSource) 
	
		::DeregisterEventSource(m_hEventSource);
}

// Default command line argument parsing

// Returns TRUE if it found an arg it recognised, FALSE if not. Note: processing 
// some arguments causes output to stdout to be generated.

void do_help(void)
{
	printf ("ECDL2K-108 95/98/NT Service available commands:\n");
	printf ("-h : this help screen\n");
	printf ("-i : install client as service on Win95/98/NT\n");
	printf ("-r : run as normal client in foreground\n");
	printf ("-u : uninstall client from service on Win95/98/NT\n");
	printf ("-v : display version info and install status\n");
	exit(1);
}

BOOL CNTService::ParseStandardArgs(int argc, char* argv[])
{	OsInfo	os;
    
    if(argc <= 1)	// See if we have any command line args we recognise
	{
		if(IsInstalled())
		
			printf("Attempting to start service...\n");
	
		else

			do_help();

		return FALSE;
    }

    if(_stricmp(argv[1], "-v") == 0)		// output version information
	{
		printf("%s Version %s\n", m_szServiceName, VERSION);

        if(os.id() == OsInfo::WinNT)

			printf("The service is %s installed\n", IsInstalled() ? "currently" : "not");
        
		return TRUE;	// say we processed the argument
    } 
	else if(_stricmp(argv[1], "-i") == 0)	// install service
	{
        if(os.id() == OsInfo::WinNT)
		{       
			if(IsInstalled()) 
			
				printf("%s is already installed\n", m_szServiceName);
              
			else if(Install())	// Try and install the copy that's running
				  
				printf("%s installed\n", m_szServiceName);
                   
			else

				printf("%s failed to install. Error %d\n", m_szServiceName, GetLastError());
		}
		else if(os.id() == OsInfo::Win95 || os.id() == OsInfo::Win98)
		{   
            printf("Installation as Windows 9x Service\n"); install9x();
		}
		else
             
			printf("This Operating System is NOT supported (sorry)\n");
 
        return TRUE;	// say we processed the argument
    } 
	else if(_stricmp(argv[1], "-u") == 0)	// uninstall service
	{
		if(os.id() == OsInfo::WinNT)
		{
			if(!IsInstalled()) 
			
				printf("%s is not installed\n", m_szServiceName);
			 
			else if(Uninstall())	// Try and remove the copy that's installed
			{	
				char szFilePath[_MAX_PATH];
				::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));
				printf("%s removed. (You must delete the file (%s) yourself.)\n",
					   m_szServiceName, szFilePath);
			} 
			else 
				
				printf("Could not remove %s. Error %d\n", m_szServiceName, GetLastError());
		}
		else if(os.id() == OsInfo::Win95 || os.id() == OsInfo::Win98)
		{        
            printf("Removing Windows 9x Service\n"); remove9x();
		}
		else
        
			printf("This Operating System is NOT supported (sorry)\n");

        return TRUE; // say we processed the argument
	} 
	else if(_stricmp(argv[1], "-r") == 0)	// run manually
	{  
        isManual = TRUE; printf("Is running as manual console\n"); 
    } 
	else if(_stricmp(argv[1], "-9") == 0)	// run a service on Windows 9x
	{    
        isManual = TRUE; FreeConsole();
		
		HMODULE hModule = ::GetModuleHandle(TEXT("kernel32.dll")); 
        fncptr = (fp_RegServProc)::GetProcAddress(hModule, "RegisterServiceProcess"); 
       
		if(fncptr != NULL) 
            
			(*fncptr)(0, 1); 
	} 
	else if(_stricmp(argv[1], "-h") == 0)	// help message

		do_help();

    return FALSE;
}

// Test if the service is currently installed

BOOL CNTService::IsInstalled(void)
{	SC_HANDLE	hSCM, hService;
	BOOL		bResult = FALSE;

    // Open the Service Control Manager with full access
    
	if(hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS))
	{   
		if(hService = ::OpenService(hSCM, m_szServiceName, SERVICE_QUERY_CONFIG))
		{
			::CloseServiceHandle(hService); bResult = TRUE; 
        }

        ::CloseServiceHandle(hSCM);
    }
    
	return bResult;
}

BOOL CNTService::Install(void)
{	SC_HANDLE	hSCM, hService;
	char		szFilePath[_MAX_PATH], szKey[_MAX_PATH]; 
    HKEY		hKey = NULL;

    // Open the Service Control Manager with full access
    
	if(!(hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS)))

		return FALSE;

    // Get the executable file path
    
	::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));

    // Create the service

    hService = ::CreateService(hSCM, m_szServiceName, m_szServiceName,
					SERVICE_ALL_ACCESS, SERVICE_WIN32_OWN_PROCESS, SERVICE_AUTO_START, 
					SERVICE_ERROR_NORMAL, szFilePath, NULL, NULL, NULL, NULL, NULL);
    if(!hService) 
	{
        ::CloseServiceHandle(hSCM); return FALSE;
	}

    // make registry entries to support logging messages.  Add the source name as a subkey 
	// under the Application key in the EventLog service portion of the registry.
    
	strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\");
	strcat(szKey, m_szServiceName);
    
	if(::RegCreateKey(HKEY_LOCAL_MACHINE, szKey, &hKey) != ERROR_SUCCESS) 
	{
        ::CloseServiceHandle(hService); ::CloseServiceHandle(hSCM); return FALSE;
    }

    // Add the Event ID message-file name to the 'EventMessageFile' subkey
    
	::RegSetValueEx(hKey, "EventMessageFile", 0, REG_EXPAND_SZ, 
										(CONST BYTE*)szFilePath, strlen(szFilePath) + 1);     

    // Set the supported types flags

    DWORD dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE;

    ::RegSetValueEx(hKey, "TypesSupported", 0, REG_DWORD, (CONST BYTE*)&dwData, sizeof(DWORD));
    
	::RegCloseKey(hKey);

    LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_INSTALLED, m_szServiceName);

    ::CloseServiceHandle(hService); ::CloseServiceHandle(hSCM); return TRUE;
}

BOOL CNTService::Uninstall(void)
{	SC_HANDLE	hSCM, hService;
	BOOL		bResult = FALSE;

    // Open the Service Control Manager with full access
    
	if(!(hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS)))
		
		return FALSE;

    if(hService = ::OpenService(hSCM, m_szServiceName, DELETE))
	{
        if(::DeleteService(hService)) 
		{
            LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_REMOVED, m_szServiceName); bResult = TRUE;
		} 
		else 

            LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_NOTREMOVED, m_szServiceName);

        ::CloseServiceHandle(hService);
    }
    
    ::CloseServiceHandle(hSCM); return bResult;
}

// This function makes an entry into the application event log

void CNTService::LogEvent(WORD wType, DWORD dwID, const char* psz1, const char* psz2, const char* psz3)
{	const char* ps[3];
    int			iStr = 0;
    
    ps[0] = psz1; if(ps[0] != NULL) iStr = 1;
	ps[1] = psz2; if(ps[1] != NULL) iStr = 2;
	ps[2] = psz3; if(ps[2] != NULL) iStr = 3;
 
    // Check the event source has been registered and register it if not
    
	if(!m_hEventSource) 
	
        m_hEventSource = ::RegisterEventSource(NULL, m_szServiceName); 

    if(m_hEventSource) 
	
        ::ReportEvent(m_hEventSource, wType, 0, dwID, NULL, iStr, 0, ps, NULL);
}

// Service startup and registration

BOOL CNTService::StartService(void)
{	SERVICE_TABLE_ENTRY st[] = { {m_szServiceName, ServiceMain}, {NULL, NULL} };
    
	if(isManual)
	{
		printf("Is manual\n"); m_bIsRunning = TRUE; OnInit(); Run(); return TRUE;
    }

    printf("Calling StartServiceCtrlDispatcher()\n");
    BOOL b = ::StartServiceCtrlDispatcher(st);
    printf("Returned from StartServiceCtrlDispatcher()\n");
    return b;
}

// status functions

void CNTService::SetStatus(DWORD dwState)
{
    DebugMsg("CNTService::SetStatus(%lu, %lu)", m_hServiceStatus, dwState);
    m_Status.dwCurrentState = dwState;
    ::SetServiceStatus(m_hServiceStatus, &m_Status);
}

// Service initialization

BOOL CNTService::Initialize(void)
{
    DebugMsg("Entering CNTService::Initialize()");

    SetStatus(SERVICE_START_PENDING);	// Start the initialization
    
    BOOL bResult = OnInit();			// Perform the actual initialization
    
    // Set final state

    m_Status.dwWin32ExitCode = GetLastError(); 
	m_Status.dwCheckPoint = 0; m_Status.dwWaitHint = 0;
    
	if(!bResult) 
	{
        LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_FAILEDINIT); SetStatus(SERVICE_STOPPED); return FALSE;    
    }
    
    LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_STARTED); SetStatus(SERVICE_RUNNING);

    DebugMsg("Leaving CNTService::Initialize()"); return TRUE;
}

// This function performs the main work of the service. 
// When this function returns the service has stopped.

void CNTService::Run(void)
{
    DebugMsg("Entering CNTService::Run()");

    while(m_bIsRunning) 
	{
        DebugMsg("Sleeping..."); Sleep(5000);
    }

    DebugMsg("Leaving CNTService::Run()");	// nothing more to do
}

// static member function (callback)

void CNTService::ServiceMain(DWORD dwArgc, LPTSTR* lpszArgv)
{
    m_pThis->DebugMsg("Entering CNTService::ServiceMain()");

    // Register the control request handler

    m_pThis->m_Status.dwCurrentState = SERVICE_START_PENDING;
    m_pThis->m_hServiceStatus = RegisterServiceCtrlHandler(m_pThis->m_szServiceName, Handler);
    
	if(m_pThis->m_hServiceStatus == NULL) 
	{
        m_pThis->LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_CTRLHANDLERNOTINSTALLED); return;
    }

    if(m_pThis->Initialize())	// Start the initialisation
	{	
		// Do the real work - when the Run function returns, the service has stopped.
        m_pThis->m_bIsRunning = TRUE;
        m_pThis->m_Status.dwWin32ExitCode = 0;
        m_pThis->m_Status.dwCheckPoint = 0;
        m_pThis->m_Status.dwWaitHint = 0;
        m_pThis->Run();
    }

    m_pThis->SetStatus(SERVICE_STOPPED);	// Tell the service manager we are stopped

    m_pThis->DebugMsg("Leaving CNTService::ServiceMain()");
}

// static member function (callback) to handle commands from the service control manager

void CNTService::Handler(DWORD dwOpcode)
{
    m_pThis->DebugMsg("CNTService::Handler(%lu)", dwOpcode);

    switch (dwOpcode) 
	{
    case SERVICE_CONTROL_STOP:
        m_pThis->SetStatus(SERVICE_STOP_PENDING);
        m_pThis->OnStop();
        m_pThis->m_bIsRunning = FALSE;
        m_pThis->LogEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_STOPPED);
        break;

    case SERVICE_CONTROL_PAUSE:	m_pThis->OnPause(); break;

    case SERVICE_CONTROL_CONTINUE: m_pThis->OnContinue(); break;

    case SERVICE_CONTROL_INTERROGATE: m_pThis->OnInterrogate(); break;

    case SERVICE_CONTROL_SHUTDOWN: m_pThis->OnShutdown(); break;

    default:
        if(dwOpcode < SERVICE_CONTROL_USER || !m_pThis->OnUserControl(dwOpcode))

            m_pThis->LogEvent(EVENTLOG_ERROR_TYPE, EVMSG_BADREQUEST);

		break;
    }

    // Report current status

    m_pThis->DebugMsg("Updating status (%lu, %lu)", m_pThis->m_hServiceStatus,
                       m_pThis->m_Status.dwCurrentState);
    ::SetServiceStatus(m_pThis->m_hServiceStatus, &m_pThis->m_Status);
}

// Control request handlers

BOOL CNTService::OnInit(void)		// Called when the service is first initialized
{
    DebugMsg("CNTService::OnInit()"); return TRUE;
}

void CNTService::OnStop(void)		// Called when the service control manager wants to stop the service
{
    DebugMsg("CNTService::OnStop()");
}

void CNTService::OnInterrogate(void)	// called when the service is interrogated
{
    DebugMsg("CNTService::OnInterrogate()");
}

void CNTService::OnPause(void)		// called when the service is paused
{
    DebugMsg("CNTService::OnPause()");
}

void CNTService::OnContinue(void)	// called when the service is continued
{
    DebugMsg("CNTService::OnContinue()");
}

void CNTService::OnShutdown(void)	// called when the service is shut down
{
    DebugMsg("CNTService::OnShutdown()");
}

// called when the service gets a user control message

BOOL CNTService::OnUserControl(DWORD dwOpcode)
{
    DebugMsg("CNTService::OnUserControl(%8.8lXH)", dwOpcode);
    return FALSE;	// say not handled
}

// Debugging support

void CNTService::DebugMsg(const char* pszFormat, ...)
{	char	buf[1024];
	va_list	arglist;
	
    sprintf(buf, "[%s](%lu): ", m_szServiceName, GetCurrentThreadId());

	va_start(arglist, pszFormat);
    vsprintf(buf + strlen(buf), pszFormat, arglist); 
	va_end(arglist);

    strcat(buf, "\n"); OutputDebugString(buf);
}

BOOL CNTService::isCmdLine(void)
{
	return TRUE;
}

char	*reg_key = "Software\\Microsoft\\Windows\\CurrentVersion\\RunServices";

int CNTService::install9x(void)
{	HKEY	key;
	char	name[_MAX_PATH];

	if(RegOpenKey(HKEY_LOCAL_MACHINE, reg_key, &key) == ERROR_SUCCESS)
	{
		::GetModuleFileName(NULL, name, _MAX_PATH); strcat(name, " -9");

		printf(RegSetValueEx(key, m_szServiceName, 0, REG_SZ, (unsigned char*)name, strlen(name))
				== ERROR_SUCCESS ? "Service installed\n" : "Service not registered\n");

		RegCloseKey(key); return TRUE;
	}
	else
	{	   
		// must create key since it doesn't exist
		if(::RegCreateKey(HKEY_LOCAL_MACHINE, reg_key, &key) == ERROR_SUCCESS)
		{
			::GetModuleFileName(NULL, name, _MAX_PATH); strcat(name, " -9");

			printf(RegSetValueEx(key, m_szServiceName, 0, REG_SZ, (unsigned char*)name, strlen(name))
				== ERROR_SUCCESS ? "*Service installed\n" : "Service not registered\n");

			RegCloseKey(key); return TRUE;
		}
		else
			printf("Error creating services key\n"); return FALSE;
	}
}

int CNTService::remove9x(void)
{	HKEY		key;
	char		name[_MAX_PATH];
   
	if(RegOpenKey(HKEY_LOCAL_MACHINE, reg_key, &key) == ERROR_SUCCESS) 
	{
		::GetModuleFileName(NULL, name, _MAX_PATH); strcat(name, " -9");

		printf(RegDeleteValue(key, m_szServiceName) == ERROR_SUCCESS 
						? "Service removed\n" : "Service not registered\n");

		RegCloseKey(key); return TRUE;
	}
	else
	{
		printf("Error opening services key\n"); return FALSE;
	}
}
