// SOURCE CODE HEADER
//
// ntservice.h
// Purpose: Run an application as a NT/95/98 service
// Contact: jeffg@cips.ca
// Legalese: This source code is public domain
// Based on MSDN example code by Nigel Thomson (nigelt@microsoft.com)
// Modified to also run as Win95/98 service by Steve Porter (steve@dcypher.net)
// Modified for the ECDL2K-108 challenge by Jeff Gilchrist (jeffg@cips.ca).

// Definitions for CNTService

#ifndef _NTSERVICE_H_
#define _NTSERVICE_H_

#include "ntservmsg.h" // Event message IDs

#define VERSION "1.1.4<brg,jg>"
#define V_MAJOR	1
#define V_MINOR	1
#define V_BUILD	4

#define SERVICE_CONTROL_USER 128

class CNTService
{
public:
    CNTService(const char* szServiceName);
    virtual ~CNTService(void);

    BOOL Install(void);
    BOOL Uninstall(void);
    BOOL Initialize(void);
    BOOL StartService(void);
    BOOL ParseStandardArgs(int argc, char* argv[]);
    void SetStatus(DWORD dwState);
	BOOL isCmdLine(void);
    BOOL IsInstalled(void);

    virtual void Run(void);
	virtual BOOL OnInit(void);
    virtual void OnStop(void);
    virtual void OnInterrogate(void);
    virtual void OnPause(void);
    virtual void OnContinue(void);
    virtual void OnShutdown(void);
    virtual BOOL OnUserControl(DWORD dwOpcode);

    void LogEvent(WORD wType, DWORD dwID, const char* psz1 = NULL, 
					const char* psz2 = NULL, const char* psz3 = NULL);
    void DebugMsg(const char* pszFormat, ...);
    
    // static member functions

    static void WINAPI ServiceMain(DWORD dwArgc, LPTSTR* lpszArgv);
    static void WINAPI Handler(DWORD dwOpcode);

    char m_szServiceName[64];
    int m_iMajorVersion;
    int m_iMinorVersion;
	int m_iBuildVersion;
    SERVICE_STATUS_HANDLE m_hServiceStatus;
    SERVICE_STATUS m_Status;
	BOOL isManual;
    BOOL m_bIsRunning;

     // static data - risky hack to get object ptr
	
	static CNTService *m_pThis;

private:
    HANDLE m_hEventSource;

protected:						// For Windows 9x
	int install9x(void);
public:
    int remove9x(void);
};

#endif
